# This Python file uses the following encoding: utf-8

import platform
import locale
if platform.system() == 'Windows':
    locale.setlocale(locale.LC_ALL, 'pt-BR')
    locale.setlocale(locale.LC_NUMERIC, 'pt-BR')
else:
    locale.setlocale(locale.LC_ALL, 'pt_BR')
    locale.setlocale(locale.LC_NUMERIC, 'pt_BR')

from enum import Enum
import decimal
#  IBM’s General Decimal Arithmetic Specification, The General Decimal Arithmetic Specification.
decimal.getcontext().prec = 2

from PyQt5 import QtCore,  QtGui, QtWidgets
from PyQt5.Qt     import QStandardItemModel, QStandardItem
from PyQt5.QtGui  import QColor,  QBrush
from PyQt5.QtCore import  QAbstractItemModel, QFile, QIODevice, QModelIndex, Qt, QAbstractTableModel
from PyQt5.QtWidgets import QWidget, QApplication, QTreeView, QTableView

etpDados = Enum( 'etpDados' , [ 'DATA', 'STRING', 'VALOR', 'CHKBOX' ] )
eusrRoles = Enum( 'eusrRoles' , [ 'MAXLIHAS', 'LINHATOTAL' ,'CRUDLANCAMENTO'] , start = QtCore.Qt.UserRole )

class seqId():
    id = 0
    #
    def next():
        i = seqId.id
        seqId.id += 1
        return i

class enumSLIP(Enum):

    def __new__(cls, *args ):
        obj = object.__new__(cls)
        obj._value_ = seqId.next()
        return obj

    DTAPLEFETIVA = ( 'Data ',  etpDados.DATA  )
    ORIGEM       = ( 'Origem', etpDados.STRING )
    TIPODC       = ( 'Tipo D/C', etpDados.STRING )
    CREDITO      = ( 'Crédito', etpDados.VALOR )
    DEBITO       = ( 'Débito', etpDados.VALOR )
    BALANCO      = ( 'Balanco',  etpDados.VALOR )
    OBSERVACAO   = ( 'Observacao', etpDados.STRING )
    EXCLUIR      = ( 'Excluir', etpDados.CHKBOX )

    def __init__(self, etiq , tpd ):
        self._etiq = etiq
        self._tpdados = tpd

    @property
    def cabecalho(self):
        return self._etiq

    @property
    def tipoDados(self):
        return self._tpdados


class lancamento:    
    def __init__(self, dtEfetiva, origem, valor, deb_cred , razao ):
            self._dtEfetiva = dtEfetiva
            self._origem = origem
            self._valor = valor
            self._deb_cred = deb_cred
            self._razao = razao
            self._ativo = True

    def columnCount( parent ):
        return len(enumSLIP.__members__)

    def coluna(self, col ):
        if  enumSLIP(col) == enumSLIP.DTAPLEFETIVA :
            return  self._dtEfetiva
        elif enumSLIP(col) == enumSLIP.ORIGEM :
            return  self._origem
        elif enumSLIP(col) == enumSLIP.TIPODC :
            return  self._deb_cred
        elif enumSLIP(col) == enumSLIP.CREDITO :
            return  self._valor if self._deb_cred == 'C' and self._ativo == True else 0
        elif enumSLIP(col) == enumSLIP.DEBITO :
            return  self._valor if self._deb_cred == 'D' and self._ativo == True else 0
        elif enumSLIP(col) == enumSLIP.BALANCO :
            vl = self._valor if self._ativo else 0
            return  vl if self._deb_cred == 'C' else -vl
        elif enumSLIP(col) == enumSLIP.OBSERVACAO:
            return  self._razao
        elif enumSLIP(col) == enumSLIP.EXCLUIR:
            return  self._ativo
        else:
            return QVariant()

    @property
    def valorBruto(self):
        return self._valor

    def remover(self, value ):
        self._ativo = value
        return self._ativo

    def __str__(self):
        return '\tData Efetiva: {} ; Valor : {} ; Tipo : {} ; Observacao {}'.format( self._dtEfetiva, self._valor, self._deb_cred, self._razao )

class lancamentosSLIP:
    def __init__(self , slips ):
        self._slips = []
        for dtEfetiva, origem , vl, dc, razao in slips:
            self._slips.append( lancamento(dtEfetiva, origem, vl, dc, razao ) )

    def append(self, slips):
        for dtEfetiva, origem, vl, dc, razao in slips:
            self._slips.append( lancamento(dtEfetiva, origem , vl, dc, razao ) )

    def linha(self, row):
        return(self._slips[row])

    def item(self, row, col ):
        return( self._slips[row].coluna(col) )

    def columnCount(self, parent ):
        return lancamento.columnCount( parent )

    def rowCount(self, parent ):
        return len( self._slips ) + 1

    def data(self,  index, role=QtCore.Qt.DisplayRole):
        if index.row() < len( self._slips ) :
            if role == eusrRoles.CRUDLANCAMENTO.value:
                return self._slips[index.row()]
            #
            if enumSLIP(index.column()) not in [ enumSLIP.BALANCO , enumSLIP.EXCLUIR ]:
                return self._slips[index.row()].coluna(index.column())
            elif enumSLIP(index.column()) == enumSLIP.BALANCO:
                sd = [ self._slips[I].coluna(enumSLIP.BALANCO) for I in range( index.row()+1 ) ]
                return sum(sd)
            elif enumSLIP(index.column()) == enumSLIP.EXCLUIR:
                return self._slips[index.row()]._ativo
            else:
                return QtCore.QVariant()
        else:
            # Saldo Total .,
            if enumSLIP(index.column()) == enumSLIP.BALANCO:
                return self.saldoAtual
            elif index.column() == 0 :
                return "Total"
            else :
                return ''


    def Hcabecalho(self, coluna):
        return enumSLIP(coluna).cabecalho

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.EditRole and enumSLIP(index.column()) == enumSLIP.EXCLUIR :
            return self._slips[index.row()].remover(value)
        else:
            return value

    @property
    def __len__(self):
        return len(self._slips)

    @property
    def saldoAtual(self):
        saldo = 0
        for slip in self._slips:
            saldo  +=  ( slip._valor if slip._deb_cred == 'C' else -slip._valor ) if slip._ativo else 0
        return saldo

    def __str__(self):
        return "Slips\n {} \nSaldo Total: {}\n".format( '\n'.join([ str(X) for X in self._slips ]) , self.saldoAtual )


#
#
#  https://stackoverflow.com/questions/17748546/pyqt-column-of-checkboxes-in-a-qtableview/17788371
#
#
class CheckBoxDelegate(QtWidgets.QItemDelegate):
    """
    A delegate that places a fully functioning QCheckBox cell of the column to which it's applied.
    """
    def __init__(self, parent):
        QtWidgets.QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        """
        Important, otherwise an editor is created if the user clicks in this cell.
        """
        return None

    def paint(self, painter, option, index):
        """
        Paint a checkbox without the label.
        """
        if not index.data(eusrRoles.LINHATOTAL.value) :
           self.drawCheck(painter, option, option.rect, QtCore.Qt.Unchecked if index.data() == False else QtCore.Qt.Checked)
        else:
           super().paint(painter, option, index)

    def editorEvent(self, event, model, option, index):
        '''
        Change the data in the model and the state of the checkbox
        if the user presses the left mousebutton and this cell is editable. Otherwise do nothing.
        '''
        if not int(index.flags() & QtCore.Qt.ItemIsEditable) > 0 :
            return False

        if event.type() == QtCore.QEvent.MouseButtonRelease and event.button() == QtCore.Qt.LeftButton:
            # Change the checkbox-state
            self.setModelData(None, model, index)
            return True

        return False


    def setModelData (self, editor, model, index):
        '''
        The user wanted to change the old state in the opposite.
        '''
        model.setData(index, False if index.data() == True else True, QtCore.Qt.EditRole)



class SLIP(QtCore.QObject):
    def __init__(self, cnpj , pDictArray = []  ):
        super(SLIP,self).__init__()
        self._cnpj  = cnpj
        self._ocorrencias = {}
        self.loadPyDict( pDictArray )

    def setOcorrencia(self, dtOcorrencia, slips ) :
        if dtOcorrencia in self._ocorrencias:
            self._ocorrencias[dtOcorrencia].append( slips )
        else:
            self._ocorrencias[dtOcorrencia] = lancamentosSLIP(slips)

    def loadPyDict( self, pDictArray ):
        for rOc in pDictArray:
            for oc, slips in rOc.items():
                 self.setOcorrencia( oc, slips )

    @property
    def cpfcnpj(self):
        return self._cnpj

    def ocorrencias(self):
        return self._ocorrencias.keys()

    def getOcorrencia(self, ndx):
        x = list( self._ocorrencias.keys() )
        return self._ocorrencias[x[ndx]]

    def lancamentos(self, dtOc ):
        return self._ocorrencias[dtOc]

    @property
    def saldoAtual(self):
        saldo = 0
        for oc in self._ocorrencias:
            sd = oc.saldoAtual
            slado  += sd
        return saldo

    def __str__(self):
        strx = ""
        for oc in self._ocorrencias.keys():
            strx += oc + " " + str(self._ocorrencias[oc]) + '\n'
        return "{} :\n{}".format( self._cnpj ,strx )


# Abstracao da Tabela de Lancamentos para o MVC
#
class abstractLancamentoSLIP(QAbstractTableModel):
    def __init__(self, referencia):
        super(abstractLancamentoSLIP,self).__init__()
        self._referencia = referencia
        # Marcador para determinar se precisa atualizar o DB com
        # Os lancamentos.
        # Emitir um sinal da necessidade de atualizacao .,
        #
        self._sujo = False
        self.rowColor1 = QColor(250,250,255)
        self.rowColor2 = QColor(230,230,230)
        self.rowColorVrm = QColor(255,128,128)
        self.hdrColorYellow = QColor(255,255,179)
        self.hdrColorGreen = QColor(179,255,179)

    # def index():
    # def parent():
    def rowCount(self, parent ):
         # QModelIndex &paren
         return self._referencia.rowCount(parent)

    def columnCount(self, parent):
         # QModelIndex &paren
         return self._referencia.columnCount(parent)

    def data( self, index, role ):
        if not index.isValid() or index.row() >= self._referencia.rowCount(None) :
            return ""

        if role == QtCore.Qt.DisplayRole:
            if enumSLIP(index.column()) in [ enumSLIP.CREDITO,enumSLIP.DEBITO,enumSLIP.BALANCO  ]:
                # Formatar valor
                #
                if self._referencia.data(index) != "":
                    return locale.currency(self._referencia.data(index),symbol=True, grouping=True)
                else:
                    return ""
            else:
                return  self._referencia.data(index,role)

        elif role == QtCore.Qt.TextAlignmentRole:
            if enumSLIP( index.column() ) in [ enumSLIP.CREDITO,enumSLIP.DEBITO,enumSLIP.BALANCO ]:
                    return int( QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter )
            elif enumSLIP( index.column() ) in [ enumSLIP.TIPODC ]:
                    return int( QtCore.Qt.AlignCenter | QtCore.Qt.AlignVCenter )

        elif role == QtCore.Qt.BackgroundColorRole:
            if ( enumSLIP(index.column()) == enumSLIP.BALANCO ):
               if index.row() == self._referencia.rowCount(None) -1:
                    # Linha do total.,
                    #
                    if self._referencia.data(index) != 0:
                        return self.rowColorVrm
                    else:
                        return self.hdrColorGreen
               elif self._referencia.data(self.createIndex( self._referencia.rowCount(None), enumSLIP.BALANCO.value, None )) == 0  :
                    return self.hdrColorGreen
               else:
                    return self.hdrColorYellow
            elif index.row() % 2:
                return self.rowColor1
            else:
                return self.rowColor2
        elif role == eusrRoles.CRUDLANCAMENTO.value:
            return self._referencia.data(index, role )

        elif role == eusrRoles.LINHATOTAL.value:
            return True if index.row() == self._referencia.rowCount(None) -1 else False
        #Qt::DisplayRole:
        #Qt::ToolTipRole:
        #Qt::WhatsThisRole:

    def headerData(self,  section, orientation, role):
        # Orientation = Horizontal --> section = Coluna
        # Orientation = Vertical   --> section = Lina
        #
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            return  self._referencia.Hcabecalho(section)

        elif role == QtCore.Qt.BackgroundRole and orientation == QtCore.Qt.Horizontal:
            print("Hdr BG" , section , role )
            if self._referencia.data( self.createIndex( self._referencia.rowCount(None), enumSLIP.BALANCO.value, None ) ) != 0:
                print("Yellow")
                return  QBrush( self.hdrColorYellow )
            else:
                print("Green")
                return  QBrush( self.hdrColorGreen )
        else:
            return  QtCore.QVariant()

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.EditRole and enumSLIP(index.column()) == enumSLIP.EXCLUIR :
            vl = self._referencia.setData(index, value, role )
            self.dataChanged.emit(self.createIndex(index.row(),0), self.createIndex(index.row(),self._referencia.columnCount(None)))
            self.dataChanged.emit(self.createIndex(self._referencia.rowCount(None)-1,0),self.createIndex(self._referencia.rowCount(None)-1,self._referencia.columnCount(None)))
            self.headerDataChanged.emit(QtCore.Qt.Horizontal,0,self._referencia.columnCount(None))
            return vl
        elif role == eusrRoles.CRUDLANCAMENTO.value:
            print('setData --role == eusrRoles.CRUDLANCAMENTO')
            return self._referencia.setData(index,value,role)
            pass
        else:
            return value

    # def setData(),
    # def flags():
    #    ItemIsEditable=False
    def flags(self, index):
          if enumSLIP(index.column()) == enumSLIP.EXCLUIR:
              return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled
          else:
              return QtCore.Qt.ItemIsEnabled

    # You can also reimplement headerData() and setHeaderData()

    # dataChanged()
    # headerDataChanged() signals
    # when reimplementing the setData() and setHeaderData()

    # Custom models need to create model indexes for other components to use.
    # createIndex()
    #    with suitable row and column , identifier
    # Qt::ItemDataRole.


#
class tabelaAlterada(QTableView):
    def __init__(self, SLIPS, frmLancamento ):
        super(tabelaAlterada,self).__init__()
        self._SLIPS = SLIPS
        self.wfrm = frmLancamento

    def SelecaoAlterada( self, selected, deselected ):
        # print( "Alterado " , selected )
        for I in selected.indexes():
            # print( "I : ", I.parent().row() if I.parent() != QModelIndex() else "RAIZ" , ">" , I.row() , " : " , I.column() )
            if I.parent() != QModelIndex():
                parent = list( W_slp.keys() )
                modelo = self._SLIPS[ parent[I.parent().row()] ].getOcorrencia(I.row())
                self.setModel(abstractLancamentoSLIP(modelo))
                self.setItemDelegateForColumn(enumSLIP.EXCLUIR.value,  CheckBoxDelegate(self) )
                self.resizeColumnsToContents()
                self.itemLancamentoSelecionado = self.selectionModel()
                # self.itemLancamentoSelecionado.selectionChanged.connect(self.wfrm.on_selectionChanged)
                self.doubleClicked.connect(self.wfrm.on_doubleClickChanged)

            else:
                self.setModel(None)


class frmLancamento(QWidget):
    def __init__(self):
        super(frmLancamento,self).__init__()
        lbl_DtApl = QLabel( 'Data da Efetiva Aplicacao / Parcela' )
        lbl_tpDC = QLabel( 'Tipo DC')
        lbl_Vl   = QLabel('Valor')
        lbl_obs  = QLabel('Histórico')
        self.cb_DC = QCheckBox('Débito')
        self.txt_Dt = QLineEdit()
        self.txt_Historico = QLineEdit()
        self.txt_Valor = QLineEdit()
        self.txt_obs   = QLineEdit()

        self.btnCopiar = QPushButton("Copiar Lancamento")
        self.btnCopiar.setCheckable(True)
        #btnInserir.clicked.connect(the_button_was_clicked)


        self.btnInserir = QPushButton("Inserir Lancamento")
        self.btnInserir.setCheckable(True)
        #btnInserir.clicked.connect(the_button_was_clicked)

        self.btnRemover = QPushButton("Remover")
        self.btnRemover.setCheckable(True)
        #btnLimpar.clicked.connect(the_button_was_clicked)

        self.btnEfetivar = QPushButton("Efetivar")
        self.btnEfetivar .setCheckable(True)
        #btnEfetivar .clicked.connect(the_button_was_clicked)

        self.frmLancamento = QGridLayout()

        self.frmLancamento.addWidget(lbl_DtApl , 0 , 0 )
        self.frmLancamento.addWidget(lbl_tpDC  , 0 , 1 )
        self.frmLancamento.addWidget(lbl_Vl    , 0 , 2)
        self.frmLancamento.addWidget(lbl_obs   , 0 , 3 )

        self.frmLancamento.addWidget(self.txt_Dt , 1 , 0 )
        self.frmLancamento.addWidget(self.cb_DC , 1 ,1 )
        self.frmLancamento.addWidget(self.txt_Valor , 1 , 2 )
        self.frmLancamento.addWidget(self.txt_Historico , 1 , 3 )

        self.frmLancamento.addWidget(self.btnCopiar , 2 , 0 )
        self.frmLancamento.addWidget(self.btnInserir , 2 , 1 )
        self.frmLancamento.addWidget(self.btnRemover , 2 , 2 )
        self.frmLancamento.addWidget(self.btnEfetivar , 2 , 3 )
        self.setLayout(self.frmLancamento)

    @QtCore.pyqtSlot('QItemSelection', 'QItemSelection')
    def on_selectionChanged(self, selected, deselected):
        print( "on_selectionChanged" )
        pass

    @QtCore.pyqtSlot('QModelIndex')
    def on_doubleClickChanged(self, ix):

       lancamento = ix.data(eusrRoles.CRUDLANCAMENTO.value)

       self.txt_Dt.setText( lancamento.coluna(enumSLIP.DTAPLEFETIVA ) )
       #
       # ix.data(ix.sibling(ix.row(),enumSLIP.ORIGEM), eusrRoles.CRUDLANCAMENTO)
       self.cb_DC.setCheckState( True if lancamento.coluna(enumSLIP.TIPODC) == 'D' else False )
       self.txt_Valor.setText(str( lancamento.valorBruto ) )
       self.txt_Historico.setText(lancamento.coluna(enumSLIP.OBSERVACAO) )
       # ix.data(createIndex(ix.row(),enumSLIP.EXCLUIR), eusrRoles.CRUDLANCAMENTO)

#


if __name__ == '__main__':
    #
    # [ { CONTRATO :
    #      [
    #        { dtLancMvt :
    #            [ [ DtLancEfetivo, Origem , Vl , D/C , RAZAO ]  , ... ]
    #        }  , ... {   }
    #      ]
    #    }, ... { }
    # ]
    #
    # FAKEDATA  for POC testing
    #
    testJSONString = """[
    { '00000123456789' :
            [ { '2021-03-01' :
                    [
                    [ '2021-03-01' ,  'BAIXA ' , 2000 , 'C' , 'COMPE' ] ,
                    [ '2021-03-01' ,  'BAIXA ' , 12000 , 'C',  'FINANCA' ],
                    [ '2021-02-12' , 'JUROS APURADOS' , 7000 , 'D' , 'JUROS' ] ,
                    [ '2021-01-12' , 'PROCESSO ' , 7000 , 'D' , 'JUROS' ]
                    ]
              } ,
              { '2021-03-01' :
                    [
                    [ '2021-02-01' ,  'BAIXA ' , 1000 , 'C', 'ADMSIC' ],
                    [ '2021-02-01' ,  'JUROS ', 1000 , 'D', 'ADMSIC' ]
                    ]
              } ,
              { '2021-01-05' :
                    [
                    [ '2020-11-01' ,  'BAIXA ' , 1000 , 'C', 'COMPE' ]
                    ]
              }
            ]
    } ,
    { '00001123456789' :
            [ { '2021-03-01' :
                    [
                    [ '2021-03-01' , 'BAIXA ' ,   8000 , 'C' , 'COMPE' ] ,
                    [ '2021-03-01' , 'BAIXA ' , 120000 , 'C',  'FINANCA' ],
                    [ '2021-02-12' , 'JUROS ',  70000 , 'D' , 'JUROS' ] ,
                    [ '2021-01-12' , 'PROCESSO ',  70000 , 'D' , 'JUROS' ]
                    ]
              } ,
              { '2021-03-01' :
                    [
                    [ '2021-02-01' , 'JUROS ', 10000 , 'D', 'ADMSIC' ]
                    ]
              } ,
              { '2021-01-05' :
                    [
                    [ '2020-11-01' , 'BAIXA ', 10000 , 'C', 'COMPE' ]
                    ]
              }
            ]
    }
   ]"""

    import sys
    import ast
    from pprint import pprint
    from PyQt5 import Qt, QtWidgets
    from PyQt5.QtWidgets import *

    tg = ast.literal_eval( testJSONString )
    W_slp = {}
    tmdlLancamentos = QStandardItemModel()
    tmdlLancamentos.setHorizontalHeaderItem(0,QStandardItem("CONTRATO"))
    tmdlLancamentos.setHorizontalHeaderItem(1,QStandardItem("DTREF"))

    #
    for registro in tg:
        for cnpj,lctos in registro.items():
            S = SLIP(  cnpj , lctos )
            cnpjItem = QStandardItem(cnpj)
            for L in  S.ocorrencias():
                cnpjItem.appendRow(QStandardItem(L))
            tmdlLancamentos.invisibleRootItem().appendRow(cnpjItem)
            W_slp[cnpj] = S
    app = QApplication(sys.argv)
    window = QWidget()
    window.resize( 880, 680 )
    tree = QTreeView()
    # tree.setAlternatingRowColors(true)
    # tree.setAllColumnsShowFocus(true)
    tree.setModel( tmdlLancamentos )
    tree.resizeColumnToContents ( 0 )
    tree.resizeColumnToContents ( 1 )
    wfrm = frmLancamento()
    xlist0 = tabelaAlterada(W_slp, wfrm )
    # model = QStandardItemModel()
    #
    #
    # print(  W_slp[ list( W_slp.keys() )[0] ][1] )
    # xlist0.setModel( W_slp[ list( W_slp.keys() )[0] ][1] )
    tree.selectionModel().selectionChanged.connect( xlist0.SelecaoAlterada )

    vbx = QVBoxLayout(window)
    vbx.addWidget(tree)
    vbx.addWidget(xlist0)
    #tree.resize(620, 840)
    # xlist0.resize(620, 880)
    # CRUD
    #  Label
    #  Botoes
    #  Adicionar
    #  Editar
    #  Remover
    #
    vbx.addWidget(wfrm)

    #
    window.show()

    sys.exit(app.exec_())


